/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sis;

import java.util.Arrays;

import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.SiSTypeNotFoundExeception;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.UnknownComparatorType;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.UnsetValueException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSTypes;

public class SiSCondition {
	// The parameter that should be compared
	private String parameter;

	// The comparator for the parameter and the value
	private String comparator;

	// The value to compare to either string or number
	private String value = null;
	private Double doubleValue = Double.NaN;

	// A list of valid comparators
	public static final String[] VALID_COMPARATORS = { ">", ">=", "=", "==", "<", "<=", "!=", "<>" };

	/**
	 * Create a condition that compares to a string
	 * 
	 * @param parameter
	 *            name of SiS parameter to compare to
	 * @param comparator
	 * @param value
	 * @throws SiSTypeNotFoundExeception
	 *             if unknown SiSType supplied
	 * @throws UnknownComparatorType
	 *             if invalid comparator supplied
	 */
	public SiSCondition(String parameter, String comparator, String value)
			throws SiSTypeNotFoundExeception, UnknownComparatorType {
		this.setParameter(parameter);
		this.setComparator(comparator);
		this.setValue(value);
	}

	/**
	 * Create a condition that compares to a number
	 * 
	 * @param parameter
	 *            name of SiS parameter to compare to
	 * @param comparator
	 * @param value
	 * @throws SiSTypeNotFoundExeception
	 *             if unknown SiSType supplied
	 * @throws UnknownComparatorType
	 *             if invalid comparator supplied
	 */
	public SiSCondition(String parameter, String comparator, Double value)
			throws SiSTypeNotFoundExeception, UnknownComparatorType {
		this.setParameter(parameter);
		this.setComparator(comparator);
		this.setValue(value);
	}

	/**
	 * Gets the parameter
	 * 
	 * @return the parameter
	 */
	public String getParameter() {
		return this.parameter;
	}

	/**
	 * Sets the Parameter
	 * 
	 * @param parameter
	 *            name of SiS parameter to compare to
	 * @throws SiSTypeNotFoundExeception
	 *             if unknown SiSType supplied
	 */
	private void setParameter(String parameter) throws SiSTypeNotFoundExeception {
		try {
			SiSTypes.class.getField(parameter);
			this.parameter = parameter;
		} catch (NoSuchFieldException | SecurityException e) {
			throw new SiSTypeNotFoundExeception();
		}
	}

	/**
	 * Gets the comparator
	 * 
	 * @return the comparator
	 */
	public String getComparator() {
		return this.comparator;
	}

	/**
	 * Sets the comparator
	 * 
	 * @param comparator
	 * @throws UnknownComparatorType
	 *             if invalid comparator supplied
	 */
	private void setComparator(String comparator) throws UnknownComparatorType {
		if (Arrays.asList(VALID_COMPARATORS).contains(comparator)) {
			this.comparator = comparator;
		} else {
			throw new UnknownComparatorType();
		}
	}

	/**
	 * Get the string value to compare to
	 * 
	 * @return the value
	 * @throws UnsetValueException
	 *             if value is unset
	 */
	public String getValue() throws UnsetValueException {
		if (this.doubleValue.equals(Double.NaN) && this.value != null) {
			return this.value;
		}
		throw new UnsetValueException();
	}

	/**
	 * Set the value to compare to a string
	 * 
	 * @param value
	 */
	private void setValue(String value) {
		this.value = value;
		this.doubleValue = Double.NaN;
	}

	/**
	 * Get the number value to compare to
	 * 
	 * @return the number
	 * @throws UnsetValueException
	 *             if number is unset
	 */
	public Double getDoubleValue() throws UnsetValueException {
		if (this.value == null && !this.doubleValue.equals(Double.NaN)) {
			return this.doubleValue;
		}
		throw new UnsetValueException();
	}

	/**
	 * Set the value to compare to a number
	 * 
	 * @param value
	 */
	private void setValue(Double value) {
		this.value = null;
		this.doubleValue = value;
	}
}
