/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sis;

import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;

/**
 * A description object that characterizes a given information in the SiS. This
 * is utilized for both: descriptions as well as requests of data (e.g., the
 * {@link SiSRequest} extends this interface). Both share common
 * QoS-characteristics modeled in this interface.
 * 
 * The interface supports chaining.
 * 
 * Remember to alter the clone() method, when adding new properties!
 * 
 * @author Bjoern Richerzhagen
 *
 */
public class SiSInfoProperties implements Cloneable {


	/**
	 * "null" properties.
	 */
	public static final SiSInfoProperties NONE = new SiSInfoProperties();

	/**
	 * Storage
	 */
	private long lastUpdateTimestamp = Time.getCurrentTime();
	private INodeID nodeId = null;
	private SiSScope scope = null;
	private Class<?> sourceComponent = null;
	private SiSComponent sis = null;

	public SiSInfoProperties setHost(Host host) {


		this.nodeId = host.getId();

		try {
			this.sis = host.getComponent(SiSComponent.class);
		} catch (ComponentNotAvailableException e) {
			// TODO How to handle SiS not available
			System.out.println("No Sis Available");
		}

		return this;
	}

	/**
	 * Get the last update time
	 * 
	 * @return the last update time
	 */
	public long getLastUpdateTimestamp() {
		return lastUpdateTimestamp;
	}

	/**
	 * Get the last update time to now()
	 * 
	 * @return the properties
	 */
	public SiSInfoProperties setLastUpdateTimestamp() {
		this.lastUpdateTimestamp = Time.getCurrentTime();
		return this;
	}

	/**
	 * Scope of this request
	 * 
	 * @return scope or null
	 */
	public SiSScope getScope() {
		return scope;
	}

	/**
	 * Set the scope of the respective information
	 * 
	 * @param scope
	 * @return the properties
	 */
	public SiSInfoProperties setScope(SiSScope scope) {
		assert scope == null ^ !scope.equals(SiSScope.SET) : "Scope must not be SET in InfoProperties";

		this.scope = scope;
		return this;
	}

	/**
	 * Get the this of the corresponding node
	 * 
	 * @return the SiS or null if SiS not present
	 */
	public SiSComponent getSis() {
		return this.sis;
	}

	/**
	 * Optional filter based on the source component(s) of the data (e.g., only
	 * Bypass-Data). This is NOT limited to HostComponents.
	 * 
	 * @param component
	 *            interface of the (desired) source component. This supports
	 *            type inheritance - if you specify e.g., PubSubComponent, you
	 *            will get data from any source extending that interface, so
	 *            from BypassPubSubComponent for example.
	 * 
	 * @return reference to the current properties instance to support chaining
	 */
	public <T> SiSInfoProperties setSourceComponent(Class<T> sourceComponent) {
		this.sourceComponent = sourceComponent;
		return this;
	}

	/**
	 * Source component of this information
	 * 
	 * @return source component or null
	 */
	public Class<?> getSourceComponent() {
		return sourceComponent;
	}

	public void setNodeId(INodeID nodeId) {
		this.nodeId = nodeId;
	}

	public INodeID getNodeId() {
		return this.nodeId;
	}

	@Override
	public SiSInfoProperties clone() {
		SiSInfoProperties prop = new SiSInfoProperties();
		prop.setScope(scope);
		prop.setSourceComponent(sourceComponent);
		prop.setNodeId(nodeId);
		return prop;
	}

	@Override
	public String toString() {
		return "SiSInfoProperties [" + (scope != null ? "scope=" + scope + ", " : "")
				+ (sourceComponent != null ? "sourceComponent=" + sourceComponent : "")
				+ (nodeId != null ? "nodeId=" + nodeId : "") + "]";
	}

}
