/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sis;

import java.util.HashSet;
import java.util.Set;

import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.component.sis.util.SiSCondition;

/**
 * Properties that are exclusive to requests (e.g., a tolerable timeout,
 * thresholds, etc.). This is implemented in the API to ensure compatibility
 * upon interface extensions of the request object.
 * 
 * @author Bjoern Richerzhagen
 *
 */
public class SiSRequest implements Cloneable {

	/**
	 * "null" request
	 */
	public static final SiSRequest NONE = new SiSRequest();

	/**
	 * Storage
	 */
	private Set<SiSCondition> conditions = new HashSet<>();
	private long maxInformationAge = Long.MAX_VALUE;
	private Set<INodeID> receivers = new HashSet<>();
	private SiSScope scope = null;
	private Class<?> sourceComponent = null;
	private long timeout = 0;

	/**
	 * Constructors
	 */
	public SiSRequest() {
		this.maxInformationAge = Long.MAX_VALUE;
		this.timeout = 0;
	}

	public SiSRequest(long timeout, long maxInformationAge, SiSScope scope, Class<?> sourceComponent) {

		assert scope == null || !scope.equals(SiSScope.SET) : "When SET is set to scope, receivers need to be defined.";

		this.timeout = timeout;
		this.maxInformationAge = maxInformationAge;
		this.scope = scope;
		this.sourceComponent = sourceComponent;
	}

	public SiSRequest(long timeout, long maxInformationAge, SiSScope scope, Set<INodeID> receivers,
			Class<?> sourceComponent) {

		assert !scope.equals(SiSScope.SET)
				^ receivers != null : "When receivers are defined, the scope needs to be SET.";

		this.timeout = timeout;
	}

	public SiSRequest(long timeout, long maxInformationAge, Class<?> sourceComponent, Set<INodeID> receivers) {
		this.maxInformationAge = maxInformationAge;
		this.receivers = receivers;
		this.scope = SiSScope.SET;
		this.sourceComponent = sourceComponent;
		this.timeout = timeout;
	}


	public SiSRequest(long timeout, long maxInformationAge, Class<?> sourceComponent, Set<INodeID> receivers,
			Set<SiSCondition> conditions) {
		this.conditions = conditions;
		this.maxInformationAge = maxInformationAge;
		this.receivers = receivers;
		this.scope = SiSScope.SET;
		this.sourceComponent = sourceComponent;
		this.timeout = timeout;
		this.receivers = receivers;
	}

	public SiSRequest(long timeout, long maxInformationAge, SiSScope scope, Set<INodeID> receivers,
			Class<?> sourceComponent, Set<SiSCondition> conditions) {
		this(timeout, maxInformationAge, scope, receivers, sourceComponent);

		this.conditions = conditions;
	}

	public SiSRequest(long timeout, long maxInformationAge, SiSScope scope, Class<?> sourceComponent,
			Set<SiSCondition> conditions) {
		this(timeout, maxInformationAge, scope, sourceComponent);

		this.conditions = conditions;
	}

	protected SiSRequest(SiSRequest request) {
		// TODO clone
		this.timeout = request.timeout;
		this.maxInformationAge = request.maxInformationAge;
		this.scope = request.scope;
		this.sourceComponent = request.sourceComponent;
		this.receivers = request.receivers;
		this.conditions = request.conditions;
	}

	/**
	 * Add a single condition to this request
	 * 
	 * @param condition
	 * @return the request
	 */
	public SiSRequest addCondition(SiSCondition condition) {
		if (condition != null)
			this.conditions.add(condition);
		return this;
	}

	/**
	 * Add a set of conditions to this request
	 * 
	 * @param conditions
	 * @return the request
	 */
	public SiSRequest addConditions(Set<SiSCondition> conditions) {
		if (conditions != null)
			this.conditions.addAll(conditions);
		return this;
	}

	/**
	 * Get the conditions of this request
	 * 
	 * @return
	 */
	public Set<SiSCondition> getConditions() {
		return this.conditions;
	}

	/**
	 * Overwrite conditions of this request
	 * 
	 * @param conditions
	 * @return the request
	 */
	public SiSRequest setConditions(Set<SiSCondition> conditions) {
		if (conditions != null)
			this.conditions = conditions;

		return this;
	}

	/**
	 * Get the timeout
	 * 
	 * @return the timeout
	 */
	public long getTimeout() {
		return timeout;
	}

	/**
	 * Get the maximum information age
	 * 
	 * @return the maximum information age
	 */
	public long getMaxInformationAge() {
		return maxInformationAge;
	}

	/**
	 * Set the maximum age for requested information
	 * 
	 * @param maxAge
	 * @return the request
	 */
	public SiSRequest setMaxInformationAge(long maxAge) {
		this.maxInformationAge = maxAge;
		return this;
	}

	/**
	 * Scope of this request
	 * 
	 * @return scope or null
	 */
	public SiSScope getScope() {
		return scope;
	}

	/**
	 * Set the scope of the respective information
	 * 
	 * @param scope
	 * @return the request itself
	 */
	public SiSRequest setScope(SiSScope scope) {
		this.scope = scope;
		return this;
	}

	/**
	 * Source component of this information
	 * 
	 * @return source component or null
	 */
	public Class<?> getSourceComponent() {
		return sourceComponent;
	}

	/**
	 * Optional filter based on the source component(s) of the data (e.g., only
	 * Bypass-Data). This is NOT limited to HostComponents.
	 * 
	 * @param component
	 *            interface of the (desired) source component. This supports
	 *            type inheritance - if you specify e.g., PubSubComponent, you
	 *            will get data from any source extending that interface, so
	 *            from BypassPubSubComponent for example.
	 * 
	 * @return reference to the current properties instance to support chaining
	 */
	public <T> SiSRequest setSourceComponent(Class<T> sourceComponent) {
		this.sourceComponent = sourceComponent;
		return this;
	}

	/**
	 * Add one nodeId to receivers. This sets the scope automatically to SET.
	 * 
	 * @param nodeId
	 * @return the request
	 */
	public SiSRequest addReceiver(INodeID nodeId) {
		if (this.receivers == null) {
			this.receivers = new HashSet<>();
		}
		this.scope = SiSScope.SET;
		this.receivers.add(nodeId);
		return this;
	}

	/**
	 * Add a set of nodeIds to receivers. This sets the scope automatically to SET.
	 * 
	 * @param nodeIds
	 * @return the request
	 */
	public SiSRequest addReceivers(Set<INodeID> nodeIds) {
		if (this.receivers == null) {
			this.receivers = new HashSet<>();
		}
		this.scope = SiSScope.SET;
		this.receivers.addAll(nodeIds);
		return this;
	}

	/**
	 * Get all receivers
	 * 
	 * @return the receivers
	 */
	public Set<INodeID> getReceivers() {
		return this.receivers;
	}

	/**
	 * Set the receivers by overwriting the old ones. This sets the scope
	 * automatically to SET.
	 * 
	 * @param receivers
	 *            new receivers
	 * @return the request
	 */
	public SiSRequest setReceivers(Set<INodeID> receivers) {
		if (receivers != null) {
			this.scope = SiSScope.SET;
			this.receivers = receivers;
		}
		return this;
	}

	@Override
	public SiSRequest clone() {
		return new SiSRequest(this);
	}

	@Override
	public boolean equals(Object obj) {
		// TODO add new fields
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		SiSRequest other = (SiSRequest) obj;
		if (maxInformationAge != other.maxInformationAge)
			return false;
		// TODO: Meaning of Scope
		if (scope != other.scope)
			return false;
		if (!receivers.equals(other.receivers))
			return false;
		if (!receivers.containsAll(other.receivers) || !other.receivers.containsAll(receivers))
			return false;
		if (sourceComponent == null) {
			if (other.sourceComponent != null)
				return false;
		} else if (!sourceComponent.equals(other.sourceComponent))
			return false;
		if (timeout != other.timeout)
			return false;
		if (!receivers.equals(other.receivers))
			return false;
		return true;
	}

	@Override
	public int hashCode() {
		// TODO add new fields
		final int prime = 31;
		int result = 1;
		result = prime * result + (int) (maxInformationAge ^ (maxInformationAge >>> 32));
		result = prime * result + ((scope == null) ? 0 : scope.hashCode());
		result = prime * result + ((sourceComponent == null) ? 0 : sourceComponent.hashCode());
		result = prime * result + (int) (timeout ^ (timeout >>> 32));
		result = prime * result + ((receivers == null) ? 0 : receivers.hashCode());
		result = prime * result + ((conditions == null) ? 0 : conditions.hashCode());
		return result;
	}

	/**
	 * Has to return true, if the given src described by
	 * {@link SiSInfoProperties} satisfies the request.
	 * 
	 * @param src
	 * @return
	 */
	public boolean satisfiableBy(SiSInfoProperties src) {
		// TODO Finish here

		/**
		 * Check max info age
		 */
		if (src.getLastUpdateTimestamp() < (Time.getCurrentTime() - getMaxInformationAge())) {
			return false;
		}

		/**
		 * Check src component
		 */
		if (getSourceComponent() != null) {
			if (src.getSourceComponent() == null) {
				return false;
			}
			if (!getSourceComponent().isAssignableFrom(src.getSourceComponent())) {
				return false;
			}
		}

		/**
		 * Check scope
		 */
		// Request Scope is null -> accept everything
		if (getScope() != null) {

			// Request scope is set, but info has no scope
			if (src.getScope() == null) {
				return false;
			}

			// Request scope is SET
			if (getScope().equals(SiSScope.SET)) {

				// info must be NODE_LOCAL
				if (!src.getScope().equals(SiSScope.NODE_LOCAL)) {
					return false;
				}

				// nodeid must not be null
				if (src.getNodeId() == null) {
					return false;
				}

				// and nodeId has to be in the receivers set
				if (!getReceivers().contains(src.getNodeId())) {
					return false;
				}

				// scope is set and not SET so info's scope has to match the
				// request's
			} else if (!getScope().equals(src.getScope())) {
				return false;
			}
		}

		/**
		 * Check conditions
		 */
		if (!getConditions().isEmpty()) {
			// When checking conditions info properties need a SiS
			if (src.getSis() == null)
				return false;

			// Evaluate each condition
			for (SiSCondition cond : getConditions()) {
				if (!cond.matches(src.getSis()))
					return false;
			}
		}

		return true;
	}
}
