/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sis.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.HashSet;
import java.util.Random;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSInfoProperties;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSRequest;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSScope;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.SiSTypeNotFoundException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.UnknownComparator;
import de.tudarmstadt.maki.simonstrator.api.component.sis.util.SiSCondition;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class SiSRequestTest {

	Random rand = new Random();

	@BeforeClass
	public static void setUpEnvironment() {

	}

	@Before
	public void setUp() {
		//
	}

	@After
	public void tearDown() {
		//
	}

	@AfterClass
	public static void tearDownEnvironment() {
		//
	}

	/**
	 * Null request
	 */
	@Test
	public void test001_nullRequest() {

		SiSRequest request = SiSRequest.NONE;

		assertEquals(0L, request.getTimeout());
		assertEquals(Long.MAX_VALUE, request.getMaxInformationAge());
		assertNull(request.getScope());
		assertNull(request.getSourceComponent());
		assertEquals(0, request.getReceivers().size());
		assertEquals(0, request.getConditions().size());
	}

	/**
	 * Simple request creation without conditions and receivers
	 */
	@Test
	public void test002_creationSimple() {

		long timeout = rand.nextLong();
		long infoAge = rand.nextLong();

		SiSRequest request = new SiSRequest(timeout, infoAge, SiSScope.NODE_LOCAL, this.getClass());

		assertEquals(timeout, request.getTimeout());
		assertEquals(infoAge, request.getMaxInformationAge());
		assertEquals(SiSScope.NODE_LOCAL, request.getScope());
		assertEquals(this.getClass(), request.getSourceComponent());
		assertEquals(0, request.getReceivers().size());
		assertEquals(0, request.getConditions().size());
	}

	/**
	 * Request creation with receivers
	 */
	@Test
	public void test003_creationReceivers() {

		long timeout = rand.nextLong();
		long infoAge = rand.nextLong();

		HashSet<INodeID> receivers = new HashSet<>();
		receivers.add(INodeID.get("A"));

		SiSRequest request = new SiSRequest(timeout, infoAge, this.getClass(), new HashSet<>(receivers));

		assertEquals(timeout, request.getTimeout());
		assertEquals(infoAge, request.getMaxInformationAge());
		assertEquals(SiSScope.SET, request.getScope());
		assertEquals(this.getClass(), request.getSourceComponent());
		assertEquals(1, request.getReceivers().size());
		assertEquals(0, request.getConditions().size());
		assertEquals(receivers, request.getReceivers());

		// Add one receiver
		receivers.add(INodeID.get("B"));
		request.addReceiver(INodeID.get("B"));

		assertEquals(2, request.getReceivers().size());
		assertEquals(receivers, request.getReceivers());

		// Add multiple receivers
		HashSet<INodeID> new_receivers = new HashSet<>();
		new_receivers.add(INodeID.get("C"));
		new_receivers.add(INodeID.get("D"));

		receivers.addAll(new_receivers);
		request.addReceivers(new_receivers);

		assertEquals(4, request.getReceivers().size());
		assertEquals(receivers, request.getReceivers());

		// Override receivers
		receivers = new HashSet<>();
		receivers.add(INodeID.get("E"));
		receivers.add(INodeID.get("F"));

		request.setReceivers(receivers);

		assertEquals(2, request.getReceivers().size());
		assertEquals(receivers, request.getReceivers());
	}

	/**
	 * Request creation with conditions
	 */
	@Test
	public void test004_creationConditions() {

		long timeout = rand.nextLong();
		long infoAge = rand.nextLong();

		HashSet<SiSCondition> conditions = new HashSet<>();

		try {
			conditions.add(new SiSCondition("PHY_SPEED", ">", 10.));
			conditions.add(new SiSCondition("ENERGY_BATTERY_CAPACITY", ">=", 0.5));
		} catch (SiSTypeNotFoundException | UnknownComparator e) {
			// TODO: Let test fail correctly
			assertTrue(false);
		}

		SiSRequest request = new SiSRequest(timeout, infoAge, SiSScope.GLOBAL, this.getClass(),
				new HashSet<>(conditions));

		assertEquals(timeout, request.getTimeout());
		assertEquals(infoAge, request.getMaxInformationAge());
		assertEquals(SiSScope.GLOBAL, request.getScope());
		assertEquals(this.getClass(), request.getSourceComponent());
		assertEquals(0, request.getReceivers().size());
		assertEquals(2, request.getConditions().size());
		assertEquals(conditions, request.getConditions());

		// Add one condition
		try {
			conditions.add(new SiSCondition("PHY_DISTANCE", "==", 23.42));
			request.addCondition(new SiSCondition("PHY_DISTANCE", "==", 23.42));
		} catch (SiSTypeNotFoundException | UnknownComparator e) {
			// TODO: Let test fail correctly
			assertTrue(false);
		}

		assertEquals(3, request.getConditions().size());
		assertEquals(conditions, request.getConditions());

		// Add multiple conditions
		try {
			HashSet<SiSCondition> new_conditions = new HashSet<>();
			new_conditions.add(new SiSCondition("ENERGY_BATTERY_LEVEL", ">", 20.0));
			new_conditions.add(new SiSCondition("LATENCY_CELL", "<", 10.0));

			conditions.addAll(new_conditions);
			request.addConditions(new_conditions);
		} catch (SiSTypeNotFoundException | UnknownComparator e) {
			// TODO: Let test fail correctly
			assertTrue(false);
		}

		assertEquals(5, request.getConditions().size());
		assertEquals(conditions, request.getConditions());
	}

	/**
	 * null satisfiable test
	 */
	@Test
	public void test005_satisfiableScopeNull() {

		SiSRequest req = new SiSRequest();

		// Scope should be ignored
		SiSInfoProperties prop = new SiSInfoProperties();
		assertTrue(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.NODE_LOCAL);
		assertTrue(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.GLOBAL);
		assertTrue(req.satisfiableBy(prop));
	}

	/**
	 * local satisfiable test
	 */
	@Test
	public void test006_satisfiableScopeLocal() {

		SiSRequest req = new SiSRequest().setScope(SiSScope.NODE_LOCAL);

		SiSInfoProperties prop = new SiSInfoProperties();
		assertFalse(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.NODE_LOCAL);
		assertTrue(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.GLOBAL);
		assertFalse(req.satisfiableBy(prop));
	}

	/**
	 * global satisfiable test
	 */
	@Test
	public void test007_satisfiableScopeGlobal() {

		SiSRequest req = new SiSRequest().setScope(SiSScope.GLOBAL);

		SiSInfoProperties prop = new SiSInfoProperties();
		assertFalse(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.NODE_LOCAL);
		assertFalse(req.satisfiableBy(prop));

		prop = prop.setScope(SiSScope.GLOBAL);
		assertTrue(req.satisfiableBy(prop));
	}

	// TODO Add test for information age

	/**
	 * source comp tests
	 */
	@Test
	public void test008_satisfiableSourceComp() {
		SiSRequest req = new SiSRequest();

		SiSInfoProperties prop = new SiSInfoProperties().setSourceComponent(this.getClass());
		assertTrue(req.satisfiableBy(prop));

		req = req.setSourceComponent(this.getClass());
		prop = prop.setSourceComponent(this.getClass());
		assertTrue(req.satisfiableBy(prop));

		prop = prop.setSourceComponent(null);
		assertFalse(req.satisfiableBy(prop));

		req = req.setSourceComponent(new SiSRequestTestUtils.TestClassScope().getClass());
		prop = prop.setSourceComponent(new SiSRequestTestUtils.TestSubClassScope().getClass());
		assertTrue(req.satisfiableBy(prop));

		req = req.setSourceComponent(new SiSRequestTestUtils.TestSubClassScope().getClass());
		prop = prop.setSourceComponent(new SiSRequestTestUtils.TestClassScope().getClass());
		assertFalse(req.satisfiableBy(prop));
	}

	@Test
	public void test009_satisfiableScopeSet() {

		MinimalSiSHost host1 = new MinimalSiSHost(INodeID.get("A"));
		SiSInfoProperties info1 = new SiSInfoProperties().setHost(host1).setScope(SiSScope.NODE_LOCAL);

		MinimalSiSHost host2 = new MinimalSiSHost(INodeID.get("B"));
		SiSInfoProperties info2 = new SiSInfoProperties().setHost(host2).setScope(SiSScope.NODE_LOCAL);

		MinimalSiSHost host3 = new MinimalSiSHost(INodeID.get("C"));
		SiSInfoProperties info3 = new SiSInfoProperties().setHost(host3).setScope(SiSScope.NODE_LOCAL);

		MinimalSiSHost host4 = new MinimalSiSHost(INodeID.get("A"));
		SiSInfoProperties info4 = new SiSInfoProperties().setHost(host4).setScope(SiSScope.GLOBAL);

		MinimalSiSHost host5 = new MinimalSiSHost(INodeID.get("C"));
		SiSInfoProperties info5 = new SiSInfoProperties().setHost(host5).setScope(SiSScope.GLOBAL);

		HashSet<INodeID> recv = new HashSet<>();
		recv.add(INodeID.get("A"));
		recv.add(INodeID.get("B"));

		SiSRequest req = new SiSRequest().setReceivers(recv);

		assertTrue(req.satisfiableBy(info1));
		assertTrue(req.satisfiableBy(info2));
		assertFalse(req.satisfiableBy(info3));
		assertFalse(req.satisfiableBy(info4));
		assertFalse(req.satisfiableBy(info5));
	}

	@Test
	public void test010_satisfiableConditionsSimple() {

		// Setup Host and Props
		MinimalSiSHost host = new MinimalSiSHost(INodeID.get("A"));
		SiSInfoProperties info = new SiSInfoProperties().setHost(host).setScope(SiSScope.NODE_LOCAL);

		try {
			SiSRequest req1 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "==", 0.0));
			assertTrue(req1.satisfiableBy(info));

			SiSRequest req2 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "==", 45.0));
			assertFalse(req2.satisfiableBy(info));

			SiSRequest req3 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "!=", 42.0));
			assertTrue(req3.satisfiableBy(info));

			SiSRequest req4 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "!=", 0.0));
			assertFalse(req4.satisfiableBy(info));

			SiSRequest req5 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "<", 42.0));
			assertTrue(req5.satisfiableBy(info));

			SiSRequest req6 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", ">", -1.0));
			assertTrue(req6.satisfiableBy(info));

			SiSRequest req7 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "<=", -23.0));
			assertFalse(req7.satisfiableBy(info));

			SiSRequest req8 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", ">=", 100.0));
			assertFalse(req8.satisfiableBy(info));

			SiSRequest req9 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", ">=", 0.0));
			assertTrue(req9.satisfiableBy(info));

			SiSRequest req10 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "<=", 0.0));
			assertTrue(req10.satisfiableBy(info));

			SiSRequest req11 = new SiSRequest()
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", ">=", 10000.0));
			assertTrue(req11.satisfiableBy(info));

			SiSRequest req12 = new SiSRequest().addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<", 10000.0));
			assertFalse(req12.satisfiableBy(info));

			SiSRequest req13 = new SiSRequest()
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<=", 20000.0));
			assertTrue(req13.satisfiableBy(info));

			SiSRequest req14 = new SiSRequest().addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", ">", 20000.0));
			assertFalse(req14.satisfiableBy(info));

			SiSRequest req15 = new SiSRequest().addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "!=", 2342.0));
			assertTrue(req15.satisfiableBy(info));

		} catch (SiSTypeNotFoundException | UnknownComparator e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	@Test
	public void test011_satisfiableConditionsComplex() {
		// Setup Host and Props
		MinimalSiSHost host = new MinimalSiSHost(INodeID.get("A"));
		SiSInfoProperties info = new SiSInfoProperties().setHost(host).setScope(SiSScope.NODE_LOCAL);

		try {
			SiSRequest req1 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "==", 0.0))
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<=", 20000.0));
			assertTrue(req1.satisfiableBy(info));

			SiSRequest req2 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "!=", 0.0))
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<=", 20000.0));
			assertFalse(req2.satisfiableBy(info));

			SiSRequest req3 = new SiSRequest().addCondition(new SiSCondition("LATENCY_CELL", "!=", 0.0))
					.addCondition(new SiSCondition("LATENCY_CELL", "==", 0.0));
			assertFalse(req3.satisfiableBy(info));

			SiSRequest req4 = new SiSRequest().addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", ">=", 10000.0))
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<=", 20000.0));
			assertTrue(req4.satisfiableBy(info));

			SiSRequest req5 = new SiSRequest().addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", ">=", 10000.0))
					.addCondition(new SiSCondition("ENERGY_BATTERY_CAPACITY", "<=", 20000.0))
					.addCondition(new SiSCondition("LATENCY_CELL", "==", 0.0));
			assertTrue(req5.satisfiableBy(info));
		} catch (SiSTypeNotFoundException | UnknownComparator e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
}
