/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.sis.util;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.Arrays;

import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSRequest;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.InformationNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.SiSTypeNotFoundException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.UnknownComparator;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSType;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSTypes;

/**
 * A generic condition that can be matched to a host
 * 
 * @author Marc Schiller
 */
public class SiSCondition {

	/*
	 * A list of valid comparators
	 */
	private static final String[] VALID_STRING_COMPARATORS = { "==", "!=" };
	private static final String[] VALID_NUMBER_COMPARATORS = { ">", ">=", "==", "<", "<=", "!=" };

	/**
	 * Storage
	 */
	private String comparator;
	private String parameter;
	private String value = null;
	private Double doubleValue = Double.NaN;

	/**
	 * Constructor
	 * 
	 * @param parameter
	 *            the SiS Parameter to match to, see {@link SiSTypes}
	 * @param comparator
	 *            how to match this condition
	 * @param value
	 *            the value to compare the SiSType to
	 * @throws SiSTypeNotFoundException
	 *             when an unknown SiSType is supplied
	 * @throws UnknownComparator
	 *             when an unknown comparator is supplied
	 */
	public SiSCondition(String parameter, String comparator, String value)
			throws SiSTypeNotFoundException, UnknownComparator {
		this.setParameter(parameter);
		this.setComparator(comparator, false);
		this.value = value;
	}

	/**
	 * Constructor
	 * 
	 * @param parameter
	 *            the SiS Parameter to match to, see {@link SiSTypes}
	 * @param comparator
	 *            how to match this condition
	 * @param value
	 *            the value to compare the SiSType to
	 * @throws SiSTypeNotFoundException
	 *             when an unknown SiSType is supplied
	 * @throws UnknownComparator
	 *             when an unknown comparator is supplied
	 */
	public SiSCondition(String parameter, String comparator, Double value)
			throws SiSTypeNotFoundException, UnknownComparator {
		this.setParameter(parameter);
		this.setComparator(comparator, true);
		this.doubleValue = value;
	}

	/**
	 * Get the the name of the SiSType to compare to
	 */
	public String getParameter() {
		return this.parameter;
	}

	/**
	 * Check and set the name of the SiSType
	 * 
	 * @param parameter
	 *            the name of the SiSType
	 * @throws SiSTypeNotFoundException
	 *             when an unknown SiSType is supplied
	 */
	private void setParameter(String parameter) throws SiSTypeNotFoundException {
		// Check if the wanted SiSType is exposed as a public field in SiSTypes
		for (Field field : SiSTypes.class.getFields()) {
			if (field.getName().equals(parameter)) {
				this.parameter = parameter;
				return;
			}
		}
		throw new SiSTypeNotFoundException();
	}

	/**
	 * Get the comparator
	 * 
	 * @return the comparator
	 */
	public String getComparator() {
		return this.comparator;
	}

	/**
	 * Check and set the comparator
	 * 
	 * @param comparator
	 * @throws UnknownComparator
	 *             when an unknown comparator is supplied
	 */
	private void setComparator(String comparator, Boolean isNumber) throws UnknownComparator {
		if (isNumber && Arrays.asList(VALID_NUMBER_COMPARATORS).contains(comparator)) {
			this.comparator = comparator;
			return;
		}

		if (!isNumber && Arrays.asList(VALID_STRING_COMPARATORS).contains(comparator)) {
			return;
		}

		throw new UnknownComparator();
	}

	/**
	 * Get the value to compare to
	 * 
	 * @return the value, if comparing to string NaN
	 */
	public Double getDoubleValue() {
		return this.doubleValue;
	}

	/**
	 * Get the value to compare to
	 * 
	 * @return the value, if comparing to number null
	 */
	public String getValue() {
		return this.value;
	}

	/**
	 * Get the generic type of SiSType field
	 * 
	 * @param the
	 *            name of the SiSType to fetch
	 * @return the Class of the generic SiSType or null if the SiSTye doesn't
	 *         exist
	 */
	private Class<?> getGenericTypeFromName(String sisType) {
		try {
			// Reflection magic
			return (Class<?>) ((ParameterizedType) SiSTypes.class.getField(sisType).getGenericType())
					.getActualTypeArguments()[0];
		} catch (NoSuchFieldException | SecurityException e) {
			return null;
		}
	}

	/**
	 * Get the SiSType Object from SiSTypes by a String name
	 * 
	 * @param sistype
	 *            the name of the SiSType to fetch
	 * @return the SiSType or null if it doesn't exist
	 */
	private SiSType<?> getSiSTypeFromString(String sistype) {
		try {
			// More reflection magic
			return (SiSType<?>) SiSTypes.class.getField(sistype).get(SiSTypes.class);
		} catch (NoSuchFieldException | SecurityException | IllegalArgumentException | IllegalAccessException e) {
			// Requested field does not exist or access is not allowed (e.g
			// private field)
			return null;
		}
	}

	public Boolean matches(SiSComponent sis) {

		// Fetch the SiSType that should be compared
		SiSType<?> sisType = getSiSTypeFromString(getParameter());

		if (sisType == null)
			return false;

		// Get the return type of the attribute
		Class<?> typeOfSis = getGenericTypeFromName(getParameter());

		if (typeOfSis == null) {
			return false;
		}

		// Fetch value of sisType from given SiS
		Object obj;
		try {
			obj = sis.get().localState(sisType, SiSRequest.NONE);
		} catch (InformationNotAvailableException e) {
			return false;
		}

		if (typeOfSis.equals(Double.class)) {
			// Compare Double
			Double dbl = (Double) obj;
			int result = Double.compare(dbl, getDoubleValue());

			switch (getComparator()) {
			case "==":
				return (result == 0);
			case ">":
				return (result > 0);
			case ">=":
				return (result >= 0);
			case "<":
				return (result < 0);
			case "<=":
				return (result <= 0);
			case "!=":
				return (result != 0);
			default:
				return false;
			}

		} else if (typeOfSis.equals(String.class)) {
			// Compare String
			String str = (String) obj;

			switch (getComparator()) {
			case "==":
				return str.equals(getValue());
			case "!=":
				return !str.equals(getValue());
			default:
				return false;
			}
		}
		// TODO Add more comparisons

		return false;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if ((obj == null) || (obj.getClass() != this.getClass()))
			return false;

		SiSCondition other = (SiSCondition) obj;

		if (!other.getParameter().equals(this.getParameter())) {
			return false;
		}

		if (other.getValue() == null && this.getValue() == null) {
			// Check double
			if (!other.getDoubleValue().equals(this.getDoubleValue())) {
				return false;
			}
		} else if (other.getDoubleValue() == Double.NaN && this.getDoubleValue() == Double.NaN) {
			// Check string
			if (!other.getValue().equals(this.getValue())) {
				return false;
			}
		} else {
			return false;
		}

		if (!other.getComparator().equals(this.getComparator())) {
			return false;
		}

		return true;
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + getComparator().hashCode();
		hash = 31 * hash + getParameter().hashCode();
		hash = 31 * hash + (getValue() == null ? 0 : getValue().hashCode());
		hash = 31 * hash + getDoubleValue().hashCode();

		return hash;
	}
}
