/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetwork;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkRoute;

public class BreadthFirstSearchRoutingAlgorithm implements RoutingAlgorithm {

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition, RoadNetworkEdge pDestination,
			List<RoadNetworkRoute> pKnownRoutes) {
		return findRoute(pNetwork, pCurrentPosition, pDestination, pKnownRoutes, new ArrayList<>());
	}

	@Override
	public RoadNetworkRoute findRoute(RoadNetwork pNetwork, RoadNetworkEdge pCurrentPosition, RoadNetworkEdge pDestination,
			List<RoadNetworkRoute> pKnownRoutes, List<RoadNetworkEdge> pEdgesToAvoid) {
		List<RoadNetworkEdge> visitedEdges = new ArrayList<>();
		Queue<PathInformation> buffer = new LinkedList<>();
		PathInformation rootPathInformation = new PathInformation(pCurrentPosition, null);
		buffer.add(rootPathInformation);
		visitedEdges.add(pCurrentPosition);

		outer:while (buffer.size() > 0) {
			PathInformation pathInformation = buffer.poll();

			RoadNetworkEdge edge = pathInformation.getEdge();

			if (pEdgesToAvoid.contains(edge)) {
				continue;
			}

			List<RoadNetworkEdge> accessibleEdges = edge.getAccessibleEdges();

			for (RoadNetworkEdge roadNetworkEdge : accessibleEdges) {
				if (roadNetworkEdge.equals(pDestination)) {
					List<RoadNetworkEdge> routeList = new ArrayList<>();
					extractRoute(routeList, new PathInformation(roadNetworkEdge, pathInformation));

					RoadNetworkRoute route = new RoadNetworkRoute(routeList);

					for (RoadNetworkEdge edgeToAvoid : pEdgesToAvoid) {
						if (route.containsEdge(edgeToAvoid)) {
							continue outer;
						}
					}

					if (!pKnownRoutes.contains(route)) {
						return route;
					}
				} else if (!visitedEdges.contains(roadNetworkEdge)) {
					PathInformation childPathInformation = new PathInformation(roadNetworkEdge, pathInformation);
					buffer.add(childPathInformation);
					visitedEdges.add(roadNetworkEdge);
				}
			}
		}
		return null;
	}

	private void extractRoute(List<RoadNetworkEdge> pEdges, PathInformation pPathInformation) {
		if (pPathInformation.getPreviousEdge() != null) {
			extractRoute(pEdges, pPathInformation.getPreviousEdge());
		}
		pEdges.add(pPathInformation.getEdge());
	}

	private class PathInformation {
		private final RoadNetworkEdge _edge;
		private final PathInformation _previousEdge;

		public PathInformation(RoadNetworkEdge pEdge,
				PathInformation pPreviousEdge) {
			_edge = pEdge;
			_previousEdge = pPreviousEdge;
		}

		public RoadNetworkEdge getEdge() {
			return _edge;
		}

		public PathInformation getPreviousEdge() {
			return _previousEdge;
		}

		@Override
		public boolean equals(Object pObj) {
			if (pObj instanceof PathInformation) {
				return ((PathInformation) pObj).getEdge().equals(_edge);
			}
			return false;
		}
	}

}
